package dev.fyloz.colorrecipesexplorer.service

import dev.fyloz.colorrecipesexplorer.config.blacklistedJwtTokens
import dev.fyloz.colorrecipesexplorer.config.defaultGroupCookieName
import dev.fyloz.colorrecipesexplorer.exception.NotFoundException
import dev.fyloz.colorrecipesexplorer.model.*
import dev.fyloz.colorrecipesexplorer.model.validation.or
import dev.fyloz.colorrecipesexplorer.repository.EmployeeGroupRepository
import dev.fyloz.colorrecipesexplorer.repository.EmployeeRepository
import org.springframework.context.annotation.Lazy
import org.springframework.security.core.userdetails.User
import org.springframework.security.core.userdetails.UserDetails
import org.springframework.security.core.userdetails.UserDetailsService
import org.springframework.security.core.userdetails.UsernameNotFoundException
import org.springframework.security.crypto.password.PasswordEncoder
import org.springframework.stereotype.Service
import org.springframework.web.util.WebUtils
import java.time.LocalDateTime
import javax.servlet.http.HttpServletRequest
import javax.servlet.http.HttpServletResponse
import javax.transaction.Transactional

interface EmployeeService : ExternalModelService<Employee, EmployeeSaveDto, EmployeeUpdateDto, EmployeeRepository> {
    /** Check if an [Employee] with the given [firstName] and [lastName] exists. */
    fun existsByFirstNameAndLastName(firstName: String, lastName: String): Boolean

    /** Gets the employee with the given [id]. */
    fun getById(id: Long, ignoreDefaultGroupUsers: Boolean, ignoreSystemUsers: Boolean): Employee

    /** Gets all employees which have the given [group]. */
    fun getByGroup(group: EmployeeGroup): Collection<Employee>

    /** Gets the default user of the given [group]. */
    fun getDefaultGroupEmployee(group: EmployeeGroup): Employee

    /** Save a default group employee for the given [group]. */
    fun saveDefaultGroupEmployee(group: EmployeeGroup)

    /** Updates de given [entity]. **/
    fun update(entity: Employee, ignoreDefaultGroupUsers: Boolean, ignoreSystemUsers: Boolean): Employee

    /** Updates the last login time of the employee with the given [employeeId]. */
    fun updateLastLoginTime(employeeId: Long, time: LocalDateTime = LocalDateTime.now()): Employee

    /** Updates the password of the employee with the given [id]. */
    fun updatePassword(id: Long, password: String): Employee

    /** Adds the given [permission] to the employee with the given [employeeId]. */
    fun addPermission(employeeId: Long, permission: EmployeePermission): Employee

    /** Removes the given [permission] from the employee with the given [employeeId]. */
    fun removePermission(employeeId: Long, permission: EmployeePermission): Employee

    /** Logout an user. Add the authorization token of the given [request] to the blacklisted tokens. */
    fun logout(request: HttpServletRequest)
}

interface EmployeeGroupService :
  ExternalNamedModelService<EmployeeGroup, EmployeeGroupSaveDto, EmployeeGroupUpdateDto, EmployeeGroupRepository> {
    /** Gets all the employees of the group with the given [id]. */
    fun getEmployeesForGroup(id: Long): Collection<Employee>

    /** Gets the default group from a cookie in the given HTTP [request]. */
    fun getRequestDefaultGroup(request: HttpServletRequest): EmployeeGroup

    /** Sets the default group cookie for the given HTTP [response]. */
    fun setResponseDefaultGroup(groupId: Long, response: HttpServletResponse)
}

interface EmployeeUserDetailsService : UserDetailsService {
    /** Loads an [User] for the given [employeeId]. */
    fun loadUserByEmployeeId(employeeId: Long, ignoreDefaultGroupUsers: Boolean = false): UserDetails
}

@Service
class EmployeeServiceImpl(
  employeeRepository: EmployeeRepository,
  @Lazy val groupService: EmployeeGroupService,
  @Lazy val passwordEncoder: PasswordEncoder,
) : AbstractExternalModelService<Employee, EmployeeSaveDto, EmployeeUpdateDto, EmployeeRepository>(employeeRepository),
  EmployeeService {
    override fun idNotFoundException(id: Long) = employeeIdNotFoundException(id)
    override fun idAlreadyExistsException(id: Long) = employeeIdAlreadyExistsException(id)

    override fun existsByFirstNameAndLastName(firstName: String, lastName: String): Boolean =
      repository.existsByFirstNameAndLastName(firstName, lastName)

    override fun getAll(): Collection<Employee> =
      super.getAll().filter { !it.isSystemUser && !it.isDefaultGroupUser }

    override fun getById(id: Long): Employee =
      getById(id, ignoreDefaultGroupUsers = true, ignoreSystemUsers = true)

    override fun getById(id: Long, ignoreDefaultGroupUsers: Boolean, ignoreSystemUsers: Boolean): Employee =
      super.getById(id).apply {
          if (ignoreSystemUsers && isSystemUser || ignoreDefaultGroupUsers && isDefaultGroupUser)
              throw idNotFoundException(id)
      }

    override fun getByGroup(group: EmployeeGroup): Collection<Employee> =
      repository.findAllByGroup(group).filter {
          !it.isSystemUser && !it.isDefaultGroupUser
      }

    override fun getDefaultGroupEmployee(group: EmployeeGroup): Employee =
      repository.findByIsDefaultGroupUserIsTrueAndGroupIs(group)

    override fun save(entity: EmployeeSaveDto): Employee =
      save(with(entity) {
          Employee(
            id,
            firstName,
            lastName,
            passwordEncoder.encode(password),
            isDefaultGroupUser = false,
            isSystemUser = false,
            group = if (groupId != null) groupService.getById(groupId) else null,
            permissions = permissions
          )
      })

    override fun save(entity: Employee): Employee {
        if (existsById(entity.id))
            throw employeeIdAlreadyExistsException(entity.id)
        if (existsByFirstNameAndLastName(entity.firstName, entity.lastName))
            throw employeeFullNameAlreadyExistsException(entity.firstName, entity.lastName)
        return super<AbstractExternalModelService>.save(entity)
    }

    override fun saveDefaultGroupEmployee(group: EmployeeGroup) {
        save(
          Employee(
            id = 1000000L + group.id!!,
            firstName = group.name,
            lastName = "EmployeeModel",
            password = passwordEncoder.encode(group.name),
            group = group,
            isDefaultGroupUser = true
          )
        )
    }

    override fun updateLastLoginTime(employeeId: Long, time: LocalDateTime): Employee {
        val employee = getById(employeeId, ignoreDefaultGroupUsers = true, ignoreSystemUsers = false)
        employee.lastLoginTime = time
        return update(
          employee,
          ignoreDefaultGroupUsers = true,
          ignoreSystemUsers = false
        )
    }

    override fun update(entity: EmployeeUpdateDto): Employee {
        val persistedEmployee by lazy { getById(entity.id) }
        return update(with(entity) {
            Employee(
              id = id,
              firstName = firstName or persistedEmployee.firstName,
              lastName = lastName or persistedEmployee.lastName,
              password = persistedEmployee.password,
              isDefaultGroupUser = false,
              isSystemUser = false,
              group = if (entity.groupId != null) groupService.getById(entity.groupId) else persistedEmployee.group,
              permissions = permissions?.toMutableSet() ?: persistedEmployee.permissions,
              lastLoginTime = persistedEmployee.lastLoginTime
            )
        })
    }

    override fun update(entity: Employee): Employee =
      update(entity, ignoreDefaultGroupUsers = true, ignoreSystemUsers = true)

    override fun update(entity: Employee, ignoreDefaultGroupUsers: Boolean, ignoreSystemUsers: Boolean): Employee {
        with(repository.findByFirstNameAndLastName(entity.firstName, entity.lastName)) {
            if (this != null && id != entity.id)
                throw employeeFullNameAlreadyExistsException(entity.firstName, entity.lastName)
        }

        return super<AbstractExternalModelService>.update(entity)
    }

    override fun updatePassword(id: Long, password: String): Employee {
        val persistedEmployee = getById(id, ignoreDefaultGroupUsers = true, ignoreSystemUsers = true)
        return super<AbstractExternalModelService>.update(with(persistedEmployee) {
            Employee(
              id,
              firstName,
              lastName,
              passwordEncoder.encode(password),
              isDefaultGroupUser,
              isSystemUser,
              group,
              permissions,
              lastLoginTime
            )
        })
    }

    override fun addPermission(employeeId: Long, permission: EmployeePermission): Employee =
      super<AbstractExternalModelService>.update(getById(employeeId).apply { permissions += permission })

    override fun removePermission(employeeId: Long, permission: EmployeePermission): Employee =
      super<AbstractExternalModelService>.update(getById(employeeId).apply { permissions -= permission })

    override fun logout(request: HttpServletRequest) {
        val authorizationCookie = WebUtils.getCookie(request, "Authorization")
        if (authorizationCookie != null) {
            val authorizationToken = authorizationCookie.value
            if (authorizationToken != null && authorizationToken.startsWith("Bearer")) {
                blacklistedJwtTokens.add(authorizationToken)
            }
        }
    }
}

const val defaultGroupCookieMaxAge = 10 * 365 * 24 * 60 * 60 // 10 ans

@Service
class EmployeeGroupServiceImpl(
  private val employeeService: EmployeeService,
  employeeGroupRepository: EmployeeGroupRepository
) : AbstractExternalNamedModelService<EmployeeGroup, EmployeeGroupSaveDto, EmployeeGroupUpdateDto, EmployeeGroupRepository>(
  employeeGroupRepository
),
  EmployeeGroupService {
    override fun idNotFoundException(id: Long) = employeeGroupIdNotFoundException(id)
    override fun idAlreadyExistsException(id: Long) = employeeGroupIdAlreadyExistsException(id)
    override fun nameNotFoundException(name: String) = employeeGroupNameNotFoundException(name)
    override fun nameAlreadyExistsException(name: String) = employeeGroupNameAlreadyExistsException(name)

    override fun existsByName(name: String): Boolean = repository.existsByName(name)
    override fun getEmployeesForGroup(id: Long): Collection<Employee> =
      employeeService.getByGroup(getById(id))

    @Transactional
    override fun save(entity: EmployeeGroup): EmployeeGroup {
        return super<AbstractExternalNamedModelService>.save(entity).apply {
            employeeService.saveDefaultGroupEmployee(this)
        }
    }

    override fun update(entity: EmployeeGroupUpdateDto): EmployeeGroup {
        val persistedGroup by lazy { getById(entity.id) }
        return update(with(entity) {
            EmployeeGroup(
              entity.id,
              if (name.isNotBlank()) entity.name else persistedGroup.name,
              if (permissions.isNotEmpty()) entity.permissions else persistedGroup.permissions
            )
        })
    }

    @Transactional
    override fun delete(entity: EmployeeGroup) {
        employeeService.delete(employeeService.getDefaultGroupEmployee(entity))
        super.delete(entity)
    }

    override fun getRequestDefaultGroup(request: HttpServletRequest): EmployeeGroup {
        val defaultGroupCookie = WebUtils.getCookie(request, defaultGroupCookieName)
          ?: throw NoDefaultGroupException()
        val defaultGroupUser = employeeService.getById(
          defaultGroupCookie.value.toLong(),
          ignoreDefaultGroupUsers = false,
          ignoreSystemUsers = true
        )
        return defaultGroupUser.group!!
    }

    override fun setResponseDefaultGroup(groupId: Long, response: HttpServletResponse) {
        val group = getById(groupId)
        val defaultGroupUser = employeeService.getDefaultGroupEmployee(group)
        response.addHeader(
          "Set-Cookie",
          "$defaultGroupCookieName=${defaultGroupUser.id}; Max-Age=${defaultGroupCookieMaxAge}; Path=/api; HttpOnly; Secure; SameSite=strict"
        )
    }
}

@Service
class EmployeeUserDetailsServiceImpl(
  private val employeeService: EmployeeService
) :
  EmployeeUserDetailsService {
    override fun loadUserByUsername(username: String): UserDetails {
        try {
            return loadUserByEmployeeId(username.toLong(), true)
        } catch (ex: NotFoundException) {
            throw UsernameNotFoundException(username)
        } catch (ex: NotFoundException) {
            throw UsernameNotFoundException(username)
        }
    }

    override fun loadUserByEmployeeId(employeeId: Long, ignoreDefaultGroupUsers: Boolean): UserDetails {
        val employee = employeeService.getById(
          employeeId,
          ignoreDefaultGroupUsers = ignoreDefaultGroupUsers,
          ignoreSystemUsers = false
        )
        return User(employee.id.toString(), employee.password, employee.authorities)
    }
}
