package dev.fyloz.colorrecipesexplorer.service

import dev.fyloz.colorrecipesexplorer.exception.RestException
import dev.fyloz.colorrecipesexplorer.model.*
import dev.fyloz.colorrecipesexplorer.service.utils.mapMayThrow
import org.springframework.http.HttpStatus
import org.springframework.stereotype.Service
import javax.transaction.Transactional

interface InventoryService {
    /** Adds each given [MaterialQuantityDto] to the inventory and returns the updated quantities. */
    fun add(materialQuantities: Collection<MaterialQuantityDto>): Collection<MaterialQuantityDto>

    /** Adds a given quantity to the given [Material]'s inventory quantity according to the given [materialQuantity] and returns the updated quantity. */
    fun add(materialQuantity: MaterialQuantityDto): Float

    /** Deducts the inventory quantity of each [Material]s in the mix according to the ratio defined in the given [mixRatio] and returns the updated quantities. */
    fun deductMix(mixRatio: MixDeductDto): Collection<MaterialQuantityDto>

    /** Deducts the inventory quantity of each given [MaterialQuantityDto] and returns the updated quantities. */
    fun deduct(materialQuantities: Collection<MaterialQuantityDto>): Collection<MaterialQuantityDto>

    /** Deducts the inventory quantity of a given [Material] by a given quantity according to the given [materialQuantity] and returns the updated quantity. */
    fun deduct(materialQuantity: MaterialQuantityDto): Float
}

@Service
class InventoryServiceImpl(
  private val materialService: MaterialService,
  private val mixService: MixService
) : InventoryService {
    @Transactional
    override fun add(materialQuantities: Collection<MaterialQuantityDto>) =
      materialQuantities.map {
          materialQuantityDto(materialId = it.material, quantity = add(it))
      }

    override fun add(materialQuantity: MaterialQuantityDto) =
      materialService.updateQuantity(
        materialService.getById(materialQuantity.material),
        materialQuantity.quantity
      )

    @Transactional
    override fun deductMix(mixRatio: MixDeductDto): Collection<MaterialQuantityDto> {
        val mix = mixService.getById(mixRatio.id)
        val firstMixMaterial = mix.mixMaterials.first()
        val adjustedFirstMaterialQuantity = firstMixMaterial.quantity * mixRatio.ratio

        fun adjustQuantity(mixMaterial: MixMaterial): Float =
          if (!mixMaterial.material.materialType!!.usePercentages)
              mixMaterial.quantity * mixRatio.ratio
          else
              (mixMaterial.quantity * adjustedFirstMaterialQuantity) / 100f

        return deduct(mix.mixMaterials.map {
            materialQuantityDto(
              materialId = it.material.id!!,
              quantity = adjustQuantity(it)
            )
        })
    }

    @Transactional
    override fun deduct(materialQuantities: Collection<MaterialQuantityDto>): Collection<MaterialQuantityDto> {
        val thrown = mutableListOf<NotEnoughInventoryException>()
        val updatedQuantities =
          materialQuantities.mapMayThrow<MaterialQuantityDto, MaterialQuantityDto, NotEnoughInventoryException>(
            { thrown.add(it) }
          ) {
              materialQuantityDto(materialId = it.material, quantity = deduct(it))
          }

        if (thrown.isNotEmpty()) {
            throw MultiplesNotEnoughInventoryException(thrown)
        }
        return updatedQuantities
    }

    override fun deduct(materialQuantity: MaterialQuantityDto): Float =
      with(materialService.getById(materialQuantity.material)) {
          if (this.inventoryQuantity >= materialQuantity.quantity) {
              materialService.updateQuantity(this, -materialQuantity.quantity)
          } else {
              throw NotEnoughInventoryException(materialQuantity.quantity, this)
          }
      }
}

class NotEnoughInventoryException(quantity: Float, material: Material) :
  RestException(
    "notenoughinventory",
    "Not enough inventory",
    HttpStatus.BAD_REQUEST,
    "Cannot deduct ${quantity}mL of ${material.name} because there is only ${material.inventoryQuantity}mL in inventory",
    mapOf(
      "material" to material.name,
      "requestQuantity" to quantity,
      "availableQuantity" to material.inventoryQuantity
    )
  )

class MultiplesNotEnoughInventoryException(exceptions: List<NotEnoughInventoryException>) :
  RestException(
    "notenoughinventory-multiple",
    "Not enough inventory",
    HttpStatus.BAD_REQUEST,
    "Cannot deduct requested quantities because there is no enough of them in inventory",
    mapOf(
      "lowQuantities" to exceptions.map { it.extensions }
    )
  )
